/* --- Constant Variables --- */
// Scroll Types
const WINDOW = 1;
const TEXTBOX = 2;

// Pref Types
const BOOL = 0;
const INT  = 1;
const CHAR = 2;

/* -------------------------- */

/* --- Global Variables ------*/
var xsCmdDispatcher = null;
var scrollTarget = null;
var scrollType = null;
var scrollbarRegExp = /^xul:(:?slider|thumb)/;
var inputboxRegExp = /^textarea|textbox/i;
var modKeys = new Array('altKey', 'ctrlKey', 'shiftKey', 'metaKey',
	'rmb', 'lmb');
var modKey = 0;
var prefs = Components.classes["@mozilla.org/preferences-service;1"].
				getService(Components.interfaces.nsIPrefBranch);
				
var xsMouseMod = -1;
var xsIsWheelLocked = false;
var xsModKey;    // modifier key 0 = alt, 1 = ctrl, 2 = shift, 3 = meta
var xsDelta;     // the amount of movements per action.
var xsDirection; // 0 = up/down => left/right, 1 = up/down => right/left
var xsSwitchTab; // 0 = disabled, 1 = active

var xsEdSwitchMode = {
	cmdArray: new Array('Normal', 'AllTags', 'HTMLSource', 'Preview'),
	NormalModeButton: 0,
	TagModeButton: 1,
	SourceModeButton: 2,
	PreviewModeButton: 3,
	switchMode: function(id) { goDoCommand('cmd_'+this.cmdArray[id]+'Mode') },
};

function xsPrefObserver()
{
	try {
		var pdi = prefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
	pdi.addObserver(this.domain, this, false);
	} catch(ex) {}
}
xsPrefObserver.prototype = {
	domain: "extensions.xscroll.",
	observe: function(subject, topic, prefName) {
		xsInit();
	}
}

window.xsPrefObserver = new xsPrefObserver();

/* --------------------------*/

function xsInit()
{
	xsModKey = getPref("extensions.xscroll.modkey", INT, 2, 0, modKeys.length-1);
	xsDelta = getPref("extensions.xscroll.delta", INT, 36, 1, 360);
	xsDirection = getPref("extensions.xscroll.direction", INT, 0, 0, 1);
	xsSwitchTab = getPref("extensions.xscroll.switchtab", BOOL, 0);

	modKey = modKeys[xsModKey];
	xsCmdDispatcher = document.commandDispatcher;
	window.addEventListener("DOMMouseScroll", xsEventMouseScroll, false);
	if (modKey == 'rmb')
		xsMouseMod = 2;
	else if (modKey == 'lmb')
		xsMouseMod = 0;

	if (xsMouseMod >= 0) {
		window.addEventListener("mousedown", xsWheelLock, false);
		window.addEventListener("mouseup", xsWheelUnlock, false);
	}
}

function getPref(aName, aType, aDefault, aMin, aMax)
{
	var retValue = null;
	switch (aType) {
	case BOOL:
		if (prefs.getPrefType(aName) == prefs.PREF_BOOL) {
			retValue = prefs.getBoolPref(aName);
		} else {
			prefs.setBoolPref(aName, aDefault);
			retValue = aDefault;
		}
		break;
	case INT:
		if (prefs.getPrefType(aName) == prefs.PREF_INT) {
			retValue = prefs.getIntPref(aName);
			if (retValue < aMin || retValue > aMax)
				retValue = aDefault;
		} else {
			prefs.setIntPref(aName, aDefault);
			retValue = aDefault;
		}
		break;
	case CHAR:
		if (prefs.getPrefType(aName) == prefs.PREF_STRING) {
			retValue = prefs.getCharPref(aName);
		} else {
			prefs.setCharPref(aName, aDefault);
			retValue = aDefault;
		}
		break;
	}
	return retValue;	
}

function xsWheelLock(aEvent)
{
	if (aEvent.button == xsMouseMod)
		xsIsWheelLocked = true;
}

function xsWheelUnlock(aEvent)
{
	xsIsWheelLocked = false;
}

function xsEventMouseScroll(aEvent)
{
	if (xsSwitchTab && aEvent.originalTarget.localName == 'tab') {
		doSwitchTab(aEvent);
	}

	if ( isXScrollable(aEvent) ) {
		doXScroll(aEvent, scrollType);
	} else if ( (xsIsWheelLocked || aEvent[modKey] || xsCmdDispatcher.focusedElement) &&
		 isTextBox(aEvent) ) {
		doXScroll(aEvent, scrollType); 
	} else if ( xsIsWheelLocked || aEvent[modKey] ) {
		scrollTarget = aEvent.originalTarget.ownerDocument.defaultView;
		scrollType = WINDOW;
		doXScroll(aEvent, scrollType); 
	}
}

function isTextBox(aEvent) {
    if (!aEvent.target.tagName ||
		(!xsIsWheelLocked || !aEvent[modKey]) &&
		 aEvent.target.tagName == 'TEXTAREA')
		return false;

    if (aEvent.target.tagName.match(inputboxRegExp)) {
		scrollTarget = xsCmdDispatcher.focusedElement;
		scrollType = TEXTBOX;
		return true;
    }
}

function isXScrollable(aEvent) {
	if (aEvent.originalTarget.tagName &&
		aEvent.originalTarget.tagName.match(scrollbarRegExp) &&
		aEvent.originalTarget.getAttribute("orient") == 'horizontal') {

		if (aEvent.target.tagName == 'TEXTAREA') {
			if (xsCmdDispatcher.focusedElement) {
				scrollTarget = xsCmdDispatcher.focusedElement;
				scrollType = TEXTBOX;
			} else return false;
		} else {
			scrollTarget = aEvent.originalTarget.ownerDocument.defaultView
			scrollType = WINDOW;
		}
		return true;
    }
	return false;
}

function doXScroll(aEvent, aType)
{
	var detail = aEvent.detail;
    if (xsDirection) detail *= -1;
	if (aType == WINDOW) {
		if (detail > 0)
			scrollTarget.scrollBy(xsDelta, 0);
		else
			scrollTarget.scrollBy(-xsDelta, 0);
    } else {
		if (detail > 0)
			scrollTarget.scrollLeft += xsDelta;
		else
			scrollTarget.scrollLeft -= xsDelta;
    }
    aEvent.preventDefault();
}

function doSwitchTab(aEvent) {
	var tabs = aEvent.originalTarget.parentNode;
	if ( window.wintype && window.wintype == 'composer:html') {
	// composer
		var id = xsEdSwitchMode[aEvent.originalTarget.id];
		var index = 0;

		for (; index < tabs.childNodes.length; index++)
			if (tabs.childNodes[index].selected)
				break;

		if (aEvent.detail > 0) {
			if (index == tabs.childNodes.length -1)
				index = -1;
			 xsEdSwitchMode.switchMode(index+1);
		} else {
			if (index == 0)
				index = tabs.childNodes.length;
			xsEdSwitchMode.switchMode(index-1);
		}
	} else if (window.windowType == "chatzilla") {
	// chatzilla
		var index = 0;
		for (i = 1; i < tabs.childNodes.length; i++) {
			if (tabs.childNodes[i].selected == true) {
				index = i;
				break;
			}
		}
		if (aEvent.detail > 0) {
			index = (index == tabs.childNodes.length-1)? 0: index;
		} else {
			index = (index == 1)? tabs.childNodes.length-2: index-2;
		}
		setCurrentObject(window.client.viewsArray[index].source);
	} else {
	// others
		tabs.advanceSelectedTab(aEvent.detail > 0 ? 1 :-1);
	}
}
